-- Base class for all crafting jobs.
local constants = require 'constants'
local job_helper = require 'jobs.job_helper'
local BaseJob = require 'jobs.base_job'

local CraftingJob = class()
radiant.mixin(CraftingJob, BaseJob)

function CraftingJob:initialize()
   self._sv.num_crafted = {}
   job_helper.initialize(self._sv)
end

--- Public functions, required for all classes
function CraftingJob:create(entity)
   job_helper.initialize(self._sv, entity)
   self:restore()
end

function CraftingJob:restore()
   self._job_component = self._sv._entity:get_component('stonehearth:job')

   --If we load and we're the current class, do these things
   if self._sv.is_current_class then
      self:_create_xp_listeners()
   end

   self.__saved_variables:mark_changed()
end

function CraftingJob:promote(json, options)  
   job_helper.promote(self._sv, json)
   local crafter_component = self._sv._entity:add_component("stonehearth:crafter")
   crafter_component:set_json(json.crafter)
   self:_create_xp_listeners()
   self.__saved_variables:mark_changed()
end

-- Returns the level the character has in this class
function CraftingJob:get_job_level()
   return self._sv.last_gained_lv
end

-- Returns whether we're at max level.
-- NOTE: If max level is nto declared, always false
function CraftingJob:is_max_level()
   return self._sv.is_max_level 
end

-- Returns all the data for all the levels
function CraftingJob:get_level_data()
   return self._sv.level_data
end

-- We keep an index of perks we've unlocked for easy lookup
function CraftingJob:unlock_perk(id)
   self._sv.attained_perks[id] = true
   self.__saved_variables:mark_changed()
end

-- Given the ID of a perk, find out if we have the perk. 
function CraftingJob:has_perk(id)
   return self._sv.attained_perks[id]
end

-- Call when it's time to level up in this class
function CraftingJob:level_up()
   job_helper.level_up(self._sv)
   self.__saved_variables:mark_changed()
end

-- Call when it's time to demote
function CraftingJob:demote()
   self:_remove_xp_listeners()

   self._sv._entity:get_component('stonehearth:crafter'):demote()
   self._sv._entity:remove_component("stonehearth:crafter")
   
   self._sv.is_current_class = false
   self.__saved_variables:mark_changed()
end

function CraftingJob:set_fine_percentage(args)
   self._sv._entity:get_component('stonehearth:crafter'):set_fine_percentage(args.percent_chance)
end

function CraftingJob:unset_fine_percentage(args)
   self._sv._entity:get_component('stonehearth:crafter'):set_fine_percentage(0)
end

--- Private functions

function CraftingJob:_create_xp_listeners()
   self._on_craft_listener = radiant.events.listen(self._sv._entity, 'stonehearth:crafter:craft_item', self, self._on_craft)
end

function CraftingJob:_remove_xp_listeners()
   if self._on_craft_listener then
      self._on_craft_listener:destroy()
      self._on_craft_listener = nil
   end
end

--When we've crafted an item, we get XP according to the min-crafter level of the item
--If there is no level specified it's a default and should get that amount
--TODO: some items should give extra bonuses if they're really cool
function CraftingJob:_on_craft(args)
   local recipe_data = args.recipe_data
   local craftable_uri = args.product_uri
   local number_crafted = self:get_number_crafted(craftable_uri)
   
   local exp_multiplier = 1
   if number_crafted == 0 then
      -- Add a curiosity multiplier
      local attributes_component = self._sv._entity:get_component('stonehearth:attributes')
      local curiosity = attributes_component:get_attribute('curiosity')
      exp_multiplier = curiosity * constants.attribute_effects.CURIOSITY_NEW_CRAFTABLE_MULTIPLER 
      if exp_multiplier < 1 then
         exp_multiplier = 1
      end
   end

   number_crafted = number_crafted + 1
   self._sv.num_crafted[craftable_uri] = number_crafted

   local level_key = 'craft_level_0'
   local level_required = 0
   if recipe_data.level_requirement then
      level_key = 'craft_level_' .. recipe_data.level_requirement
      level_required = recipe_data.level_requirement
   end
   local job_level = self:get_job_level()
   local exp = self._sv.xp_rewards[level_key]
   if level_required < job_level then
      local difference = job_level - level_required
      exp = exp / difference
   end
   assert(exp, self._sv.job_name .. ' has no exp reward tuned for level ' .. level_required .. ' recipes')
   exp = radiant.math.round(exp * exp_multiplier)
   self._job_component:add_exp(exp)
end

--Return the number of craftable_uri this crafter has crafted in their lifetime
function CraftingJob:get_number_crafted(craftable_uri)
   local number_crafted = self._sv.num_crafted[craftable_uri]
   return number_crafted or 0
end

function CraftingJob:apply_buff(args)
   radiant.entities.add_buff(self._sv._entity, args.buff_name)   
end

function CraftingJob:remove_buff(args)
   radiant.entities.remove_buff(self._sv._entity, args.buff_name)
end

return CraftingJob